! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!+EVENT
! Module: evarimgen_mod
!
! Authors:  Ian Stewart - SSC/LUX
!          
! Description: This task makes an image in which the time-variability of each pixel is colour encoded.
! 
! Comments: 
!	TODO:
!	- No protection at present against the situation in which the number of events in the table is larger than int32 can handle. However this is about 2 billion events... probably unlikely!
!	- Change the prog so that, if withimageset=T, the user cannot also specify his own values of image sizes etc: instead these are read from the supplied image.
!	- Not sure that anything useful can be done with the mask file at this point. It has been commented out for now.
!	- warning format.
! $Id: module.f90,v 1.6 1998/05/27 13:45:39 gvacanti Exp $
!- 
module evarimgen_mod

! You should write the module name in the 'use' statement, the package name
! (and version number) in the DEPEND file, and the * from the lib*.a name in
! <package>/src/Makefile.
  use dal
  use param  
  use errorHandling
!  use dss

  use sort_mod ! part of the ssclib task

  implicit none 

  ! Globals:

  real(kind=double) ::&
    pi,&
    twoPi,&
    piOn180,&
    rootPi,&
    root2,&
    root3
  real(kind=single), parameter ::&
    halfProbKs = 0.827574

! Command-line parameters:

  character(len=20) ::&
    xColName, yColName
  character(len=80) ::&
    imageBinning, xRangeType, yRangeType
  character(len=255) ::&
    eventSetName,&
    imageSetName,&
!    maskSetName,&
testType,&
    varImageSetName
  integer(kind=int16) ::&
    numXpixels,&
    numYpixels
  integer(kind=int32) ::&
    cutoff
  real(kind=single) ::&
    xMin,&
    xMax,&
    yMin,&
    yMax,&
    xPixelSize,&
    yPixelSize
  logical ::&
    withImage,&
!    withMask,&
!    squarePixels,&
    sortEvlistReq

  private 

  public evarimgen

contains 

  !---------------------------------------------------------------------
  subroutine evarimgen
  !---------------------------------------------------------------------

    implicit none

    real(kind=single), dimension(:,:), pointer ::&
      fluxImage,&
!      chi2image,&
      ksImage
    integer, dimension(:), pointer :: index
    integer(kind=int16) ::&
      xsize,&
      ysize
    type(DataSetT) ::&
      eventSet
    type(TableT)   ::&
      eventTab
    real(kind=single) ::&
      xDelta,&
      yDelta

    !-------------------------------------------------------------------
    ! Start:

    call init
    call readParameters

!call plotKsProbFn
!return

    if (testType .ne. '') then
      if (testType .eq. 'evlist') then
        call makeTestEvlist
      else if (testType .eq. 'ts1') then
        call tsCheck1
      else if (testType .eq. 'ts2') then
        call tsCheck2
      end if
  return
    end if

    call checkEvlist(eventSet, eventTab)

    call checkOrderOrSortEvlist(eventTab, index)

    call calcImageSize(eventTab, xsize, ysize, xDelta, yDelta)

    if (withImage) then
      call loadFluxImage(fluxImage, xsize, ysize)
    else
      call makeFluxImage(eventTab, fluxImage, xsize, ysize, xDelta, yDelta)
    end if

    call makeKsImage(eventTab, index, fluxImage, ksImage, xsize, ysize&
    , xDelta, yDelta)

!    call convert2chiSq(ksImage, chi2image, xsize, ysize)

!    call writeFitsOutput(chi2image, varImageSetName, eventTab, xsize, ysize)
!    deallocate(chi2image)
    call writeFitsOutput(ksImage, varImageSetName, eventTab, xsize, ysize)
    deallocate(ksImage)

    if (.not. withImage) then
      call writeFitsOutput(fluxImage, imageSetName, eventTab, xsize, ysize)
    end if

    call release(eventSet)

  end subroutine evarimgen

  !---------------------------------------------------------------------
  subroutine init
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

    pi = 4.0 * atan2(1.0, 1.0)
    twoPi = 2.0 * pi
    piOn180 = pi / 180.0
    rootPi = sqrt(pi)
    root2 = sqrt(2.0)
    root3 = sqrt(3.0)

  end subroutine init

  !---------------------------------------------------------------------
  subroutine readParameters
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

testType = stringParameter('testtype')

    xColName = stringParameter('xcolumn')
    yColName = stringParameter('ycolumn')
    xRangeType = stringParameter('xrangetype')
    if (xRangeType .eq. 'user') then
      xMin = realParameter('ximagemin')
      xMax = realParameter('ximagemax')
    end if
    yRangeType = stringParameter('yrangetype')
    if (yRangeType .eq. 'user') then
      yMin = realParameter('yimagemin')
      yMax = realParameter('yimagemax')
    end if
    imageBinning = stringParameter('imagebinning')
    if (imageBinning .eq. 'numberofpixels') then
      numXpixels = intParameter('nxpixels')
      numYpixels = intParameter('nypixels')
!      squarePixels = booleanParameter('squarepixels')

    else if (imageBinning .eq. 'pixelsize') then
      xPixelSize = realParameter('xpixelsize')
      yPixelSize = realParameter('ypixelsize')
      if (xPixelSize .le. 0.0) then
call error('', '')
!********* unnecessary after Giuseppe's promised mods to the param interface
      end if
      if (yPixelSize .le. 0.0) then
call error('', '')
!********* unnecessary after Giuseppe's promised mods to the param interface
      end if

!    else
!call error('badImageBinningType', 'The value of the parameter imagebinning is unrecognised.')
    end if
    eventSetName = stringParameter('evlistset')
    withImage = booleanParameter('withimageset')
    imageSetName = stringParameter('imageset')
!    withMask = booleanParameter('withmaskset')
!    if (withMask) then
!      maskSetName = stringParameter('maskset')
!    end if
    sortEvlistReq = booleanParameter('sortevlist')
    varImageSetName = stringParameter('varimageset')
    cutoff = intParameter('cutoff')

  end subroutine readParameters

  !---------------------------------------------------------------------
  subroutine checkEvlist(eventSet, eventTab)
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT), intent(out) ::&
      eventSet
    type(TableT), intent(out)   ::&
      eventTab

    !-------------------------------------------------------------------
    ! Start:

    eventSet = dataSet(eventSetName, READ)
    eventTab = table(eventSet, 'EVENTS')

!*****

  end subroutine checkEvlist

  !---------------------------------------------------------------------
  subroutine checkOrderOrSortEvlist(eventTab, index)
  !---------------------------------------------------------------------

    implicit none

    integer, dimension(:), pointer :: index
    type(TableT), intent(in)   ::&
      eventTab
    type(ColumnT) ::&
      tCol
    real(kind=double), dimension(:), pointer ::&
      tColData
    integer(kind=int32) ::&
      event,&
      numEvents,&
      numEventsOutOfOrder
    real(kind=single) ::&
      timeSpan,&
      deltaT,&
      maxInverseDeltaT,&
      maxFrac

    !-------------------------------------------------------------------
    ! Start:

    numEvents = numberOfRows(eventTab)
    tCol = column(eventTab, 'TIME', READ)
    call readColData(tCol, tColData)

    if (sortEvlistReq) then
      call message(AppMsg, Verbose, 'Sorting event list.')

      allocate(index(numEvents))
!      call quick_sort(tColData(0:numEvents - 1), index) ! a routine from
call quickSort(tColData(0:numEvents - 1), index) ! a routine from
      ! ssclib/sort_mod

      ! The index file is sent to makeKsImage() and, if sortEvlistReq,
      ! used there to access the event X and Y values in the correct order.

    else ! check that the events are indeed sorted in time, warn if not:
      call message(AppMsg, Verbose, 'Checking event list time-order.')

      timeSpan = real(maxval(tColData) - minval(tColData), single)
      numEventsOutOfOrder = 0
      maxInverseDeltaT = 0.0
      do event = 1, numEvents - 1
        deltaT = real(tColData(event) - tColData(event - 1), single)
        if (deltaT .lt. 0.0) then
          numEventsOutOfOrder = numEventsOutOfOrder + 1
          maxInverseDeltaT = max(maxInverseDeltaT, -deltaT) 
        end if
      end do

      if (numEventsOutOfOrder .gt. 0) then
        maxFrac = maxInverseDeltaT / timeSpan
        write(errstr, '(a, i9, a, i9, a, f5.3, a)') "There are "&
        , numEventsOutOfOrder, " out-of-order pairs of events, from a total &
        &of ", numEvents, ". The maximum reverse time separation, as a &
        &fraction of the duration between the first and last events, is "&
        , maxFrac, ". The variability statistic works best with a &
        &time-ordered event list."

        call warning('unorderedEvlist')
      end if
    end if

  end subroutine checkOrderOrSortEvlist

  !---------------------------------------------------------------------
  subroutine calcImageSize(eventTab, xsize, ysize, xDelta, yDelta)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)   ::&
      eventTab
    integer(kind=int16), intent(out) ::&
      xsize,&
      ysize
    real(kind=single), intent(out) ::&
      xDelta,&
      yDelta
    type(ColumnT) ::&
      xCol,&
      yCol

    !-------------------------------------------------------------------
    ! Start:

    xCol = column(eventTab, xColName, READ)

    select case(xRangeType)
    case('user')
    case('maxlegal')
      xMin = real64Attribute(xCol, 'TLMIN')
      xMax = real64Attribute(xCol, 'TLMAX')
    case('maxactual')
      xMin = real64Attribute(xCol, 'TDMIN')
      xMax = real64Attribute(xCol, 'TDMAX')
    case default
    end select

    call release(xCol)

    yCol = column(eventTab, yColName, READ)

    select case(yRangeType)
    case('user')
    case('maxlegal')
      yMin = real64Attribute(yCol, 'TLMIN')
      yMax = real64Attribute(yCol, 'TLMAX')
    case('maxactual')
      yMin = real64Attribute(yCol, 'TDMIN')
      yMax = real64Attribute(yCol, 'TDMAX')
    case default
    end select

    call release(yCol)

    if (imageBinning .eq. 'pixelsize') then
      if ((xMax - xMin) / xPixelSize .gt. 1000) then
write(errstr, '(a, e9.3, a, e9.3, a, e9.3, a)') "X range ", xMin, " to "&
, xMax, " divided by pixel size ", xPixelSize, " gives too many pixels &
&in the X direction! Has to be less than 1000."
call error('xPixelTooSmall')
      end if
      xsize = floor((xMax - xMin) / xPixelSize, int16)
      xMax = xMin + xsize * xPixelSize

      if ((yMax - yMin) / yPixelSize .gt. 1000) then
write(errstr, '(a, e9.3, a, e9.3, a, e9.3, a)') "Y range ", yMin, " to "&
, yMax, " divided by pixel size ", yPixelSize, " gives too many pixels &
&in the Y direction! Has to be less than 1000."
call error('yPixelTooSmall')
      end if
      ysize = floor((yMax - yMin) / yPixelSize, int16)
      yMax = yMin + ysize * yPixelSize

    else
      xsize = numXpixels
      ysize = numYpixels
      xPixelSize = (xMax - xMin) / real(xsize)
      yPixelSize = (yMax - yMin) / real(ysize)
    end if
    xDelta = nearest(xMax - xMin, 1.0) ! The 'nearest' is so events with eg x
    yDelta = nearest(yMax - yMin, 1.0) ! The 'nearest' is so events with eg x
    ! exactly = xMax nevertheless fall into the top bin - they wouldn't
    ! otherwise.

!print *, xsize, ysize
!call error('','')

  end subroutine calcImageSize

  !---------------------------------------------------------------------
  subroutine loadFluxImage(fluxImage, xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) ::&
      xsize,&
      ysize
    type(DataSetT) ::&
      inSet
    type(ArrayT) ::&
      imageArray
    integer, dimension(:), pointer ::&
      dimVector
    integer ::&
      tempXsize,&
      tempYsize
    real(kind=single), dimension(:,:), pointer ::&
      fluxImage
    integer(kind=int8), dimension(:,:), pointer ::&
      arrayDataInt8
    integer(kind=int16), dimension(:,:), pointer ::&
      arrayDataInt16
    integer(kind=int32), dimension(:,:), pointer ::&
      arrayDataInt32
    real(kind=single), dimension(:,:), pointer ::&
      arrayDataReal32
    real(kind=double), dimension(:,:), pointer ::&
      arrayDataReal64

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Loading flux image.')

    inSet = dataSet(imageSetName, READ)
    imageArray = array(inSet, 0, READ)
    dimVector => dimensions(imageArray)
    tempXsize = dimVector(0)
    tempYsize = dimVector(1)

    if (tempXsize .ne. xsize .or. tempYsize .ne. ysize) then
call error('badImageSize', 'The number of pixels in the input image is not &
&equal to the number specified for the variability image. Sorry, I know &
&this is rather braindead, I aim to fix it soon.')
!**** later warning and rebinning.
    end if

    allocate(fluxImage(xsize, ysize))
    if (dataType(imageArray) .eq. INTEGER8) then
      arrayDataInt8 => int8Array2Data(imageArray)
      fluxImage(1:xsize, 1:ysize)&
      = real(arrayDataInt8, single)
    else if (dataType(imageArray) .eq. INTEGER16) then
      arrayDataInt16 => int16Array2Data(imageArray)
      fluxImage(1:xsize, 1:ysize)&
      = real(arrayDataInt16, single)
    else if (dataType(imageArray) .eq. INTEGER32) then
      arrayDataInt32 => int32Array2Data(imageArray)
      fluxImage(1:xsize, 1:ysize)&
      = real(arrayDataInt32, single)
    else if (dataType(imageArray) .eq. REAL32) then
      arrayDataReal32 => real32Array2Data(imageArray)
      fluxImage(1:xsize, 1:ysize)&
      = arrayDataReal32
    else ! this gives a DAL error if data type is none of the above.
      arrayDataReal64 => real64Array2Data(imageArray)
      fluxImage(1:xsize, 1:ysize)&
      = real(arrayDataReal64, single)
    end if

    call release(inSet)

    if (maxval(fluxImage - floor(fluxImage)) .gt. 0.0) then
call error('scaledImage', 'The input image has been scaled or rebinned so &
&that a non-integer number of events is contained in some pixels. This &
&invalidates the image for use in making a variability plot. Rerun &
&evarimgen with withimageset=no.')
    end if

  end subroutine loadFluxImage

  !---------------------------------------------------------------------
  subroutine makeFluxImage(eventTab, fluxImage, xsize, ysize, xDelta, yDelta)
  !---------------------------------------------------------------------

    implicit none

    type(TableT), intent(in)   ::&
      eventTab
    integer(kind=int16), intent(in) ::&
      xsize,&
      ysize
    real(kind=single), intent(in) ::&
      xDelta,&
      yDelta
    real(kind=single), dimension(:,:), pointer ::&
      fluxImage
    integer(kind=int32) ::&
      event,&
      numEvents
    type(ColumnT) ::&
      xCol,&
      yCol
    real(kind=double), dimension(:), pointer ::&
      xColData,&
      yColData
    integer(kind=int16) ::&
      xi,&
      yi

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Creating flux image.')

    numEvents = numberOfRows(eventTab)
    xCol = column(eventTab, xColName, READ)
    yCol = column(eventTab, yColName, READ)
    call readColData(xCol, xColData)
    call readColData(yCol, yColData)
    allocate(fluxImage(xsize, ysize))
    fluxImage = 0.0
    do event = 0, numEvents - 1 ! Could some of this be done by array ops?
      xi = 1 + floor(((xColData(event) - xMin) / xDelta) * xsize)
      yi = 1 + floor(((yColData(event) - yMin) / yDelta) * ysize)
      if (xi .gt. 0 .and. xi .le. xsize&
      .and. yi .gt. 0 .and. yi .le. ysize) then
        fluxImage(xi, yi) = fluxImage(xi, yi) + 1.0
      end if
    end do

    call release(xCol)
    call release(yCol)

  end subroutine makeFluxImage

  !---------------------------------------------------------------------
  subroutine readColData(col, colData)
  !---------------------------------------------------------------------
  ! This sub allows the calling routine to get the data from a column of
  ! any of the types int8, int16, int32, real32 and real64 without needing
  ! to know which it is. May even expand to accommodate boolean later. The
  ! data is returned in a real64 pointer colData.

    implicit none 

    type(ColumnT), intent(in) ::&
      col
    integer(kind=int8), dimension(:), pointer ::&
      colDataInt8
    integer(kind=int16), dimension(:), pointer ::&
      colDataInt16
    integer(kind=int32), dimension(:), pointer ::&
      colDataInt32
    real(kind=single), dimension(:), pointer ::&
      colDataReal32
    real(kind=double), dimension(:), pointer ::&
      colDataReal64,&
      colData

    !-------------------------------------------------------------------
    ! Start:

!    call message(AppMsg, Verbose, 'readColData called.')

    allocate(colData(0:numberOfRows(col) - 1))

    select case(dataType(col))
    case (INTEGER8)
      colDataInt8 => int8Data(col)
      colData = real(colDataInt8, double)
    case (INTEGER16)
      colDataInt16 => int16Data(col)
      colData = real(colDataInt16, double)
    case (INTEGER32)
      colDataInt32 => int32Data(col)
      colData = real(colDataInt32, double)
    case (REAL32)
      colDataReal32 => real32Data(col)
      colData = real(colDataReal32, double)
    case default ! this gives a DAL error if data type is none of the above.
      colDataReal64 => real64Data(col)
      colData = colDataReal64
    end select

  end subroutine readColData

  !---------------------------------------------------------------------
  subroutine makeKsImage(eventTab, index, fluxImage, ksImage&
  , xsize, ysize, xDelta, yDelta)
  !---------------------------------------------------------------------
  ! This subroutine plus the following (convert2chiSq) incorporate a Kolmogorov-Smirnov procedure detailed in Giommi P, White N E and Angelini L, "A Method for Visualising Time Variability in X-Ray Images", Astronomical Data Analysis Software and Systems IV, (1994).

    implicit none

    type(TableT), intent(in) ::&
      eventTab
    integer(kind=int16), intent(in) :: xsize, ysize
    real(kind=single), intent(in) :: xDelta, yDelta
    real(kind=single), dimension(:,:), pointer :: fluxImage, ksImage
    integer, dimension(:), pointer :: index
    real(kind=single) :: sqrtNe
    type(ColumnT) ::&
      xCol, yCol
    real(kind=double), dimension(:), pointer ::&
      xColData, yColData
    integer(kind=int16) :: xi, yi
    integer(kind=int32), dimension(xsize, ysize) :: counts
    real(kind=single) :: distance, realNumEvents
    integer(kind=int32) :: event, numEvents

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Creating variability image.')

    numEvents = numberOfRows(eventTab)
    realNumEvents = real(numEvents, single)
    xCol = column(eventTab, xColName, READ)
    yCol = column(eventTab, yColName, READ)
    call readColData(xCol, xColData)
    call readColData(yCol, yColData)

    allocate(ksImage(xsize, ysize))
    ksImage = 0.0
    counts = 0
    if (sortEvlistReq) then
      do event = 1, numEvents ! Could some of this be done by array ops?
        xi = 1 + floor(((xColData(index(event) - 1) - xMin) / xDelta) * xsize)
        yi = 1 + floor(((yColData(index(event) - 1) - yMin) / yDelta) * ysize)
        if (xi .gt. 0 .and. xi .le. xsize&
        .and. yi .gt. 0 .and. yi .le. ysize&
        .and. fluxImage(xi, yi) .gt. 0.0) then
          counts(xi, yi) = counts(xi, yi) + 1
          distance = abs((event - 1) / realNumEvents - counts(xi, yi)&
          / fluxImage(xi, yi))
!******** NOTE!!!! realNumEvents is not the correct value to use if some events are chopped off by xmax, xmin. Need another pass-through of the evlist.

          if (ksImage(xi, yi) .lt. distance) ksImage(xi, yi) = distance
!if (xi .eq. 85 .and. yi .eq. 85) then
!print *, event, index(event), xColData(index(event) - 1)&
!, yColData(index(event) - 1), counts(xi, yi), event / realNumEvents&
!, counts(xi, yi) / fluxImage(xi, yi), distance, ksImage(xi, yi)
!end if
        end if
      end do

    else ! no sort requested.
      do event = 0, numEvents - 1 ! Could some of this be done by array ops?
        xi = 1 + floor(((xColData(event) - xMin) / xDelta) * xsize)
        yi = 1 + floor(((yColData(event) - yMin) / yDelta) * ysize)
        if (xi .gt. 0 .and. xi .le. xsize&
        .and. yi .gt. 0 .and. yi .le. ysize&
        .and. fluxImage(xi, yi) .gt. 0.0) then
          counts(xi, yi) = counts(xi, yi) + 1
          distance = abs(event / realNumEvents - counts(xi, yi)&
          / fluxImage(xi, yi))
!******** NOTE!!!! realNumEvents is not the correct value to use if some events are chopped off by xmax, xmin. Need another pass-through of the evlist.

          if (ksImage(xi, yi) .lt. distance) ksImage(xi, yi) = distance
!if (xi .eq. 85 .and. yi .eq. 85) then
!print *, event, xColData(event), yColData(event), counts(xi, yi)&
!, event / realNumEvents, counts(xi, yi) / fluxImage(xi, yi), distance&
!, ksImage(xi, yi)
!end if
        end if
      end do
    end if

    call release(xCol)
    call release(yCol)

    ! Now normalise ksImage:
    do xi = 1, xsize
      do yi = 1, ysize
        sqrtNe = sqrt(numEvents * (counts(xi, yi) / real(numEvents&
        + counts(xi, yi))))

        ksImage(xi, yi) = ksimage(xi, yi) * (sqrtNe + 0.12 + 0.11 / sqrtNe)
        ! This formula is found in Numerical Recipes in Fortran 77, 2nd
        ! edition. The probability of the resulting value of ksImage being
        ! 'natural' is given by
        !	   
        !	                __inf
        !	                \      j-1
        !	 P(ksImage) = 2  \ (-1)   exp[-2*(j * ksImage)^2]
        !	                 /
        !	                /_
        !	                j=1
        !

!        ksImage(xi, yi) = ksimage(xi, yi) - halfProbKs
        ksImage(xi, yi) = ksimage(xi, yi)
      end do
    end do

    ! Blank off pixels with too few counts:
!    where (counts .lt. cutoff) ksImage = 0.0
    where (counts .lt. cutoff) ksImage = halfProbKs

!    where (fluxImage .gt. 0.0) ksImage = ksImage / fluxImage
!!! Doesn't give a good effect.

  end subroutine makeKsImage

  !---------------------------------------------------------------------
  subroutine convert2chiSq(ksImage, chi2image, xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    integer(kind=int16), intent(in) :: xsize, ysize
    real(kind=single), dimension(:,:), pointer :: ksImage, chi2image

    !-------------------------------------------------------------------
    ! Start:

    allocate(chi2image(xsize, ysize))

!******

    deallocate(ksImage)

  end subroutine convert2chiSq

  !---------------------------------------------------------------------
  subroutine writeFitsOutput(image, filename, eventTab, xsize, ysize)
  !---------------------------------------------------------------------

    implicit none

    real(kind=single), dimension(:,:), pointer :: image
    integer(kind=int16), intent(in) :: xsize, ysize
    character(len=*), intent(in) :: filename
    real(kind=single) ::&
      yStretch
    type(DataSetT) :: outSet
    type(ArrayT) :: outArray
    integer, dimension(0:1) ::&
      dimVector
    real(kind=double) :: xRefPixel, yRefPixel, crpix1, crpix2, crval1, crval2, cdelt1, cdelt2
    character(len=20) :: ctype1, ctype2, cunit1, cunit2
    real(kind=single), dimension(:,:), pointer ::&
      arrayDataReal32
    type(TableT), intent(in) ::&
      eventTab
    type(ColumnT) ::&
      xCol, yCol

    !-------------------------------------------------------------------
    ! Start:

    call message(AppMsg, Verbose, 'Writing output image '&
    //trim(filename)//'.')

!    allocate(dimVector(0:1))

    outSet = dataSet(filename, CREATE)
    dimVector(0) = int(xsize) ! the 'ints' just do a conversion from the int16
    dimVector(1) = int(ysize) ! of xsize, ysize to the generic int of dimVector.
yStretch = 1.0

    xCol = column(eventTab, xColName, READ)
    if (.not. (hasAttribute(xCol, 'TCTYP')&
    .and. hasAttribute(xCol, 'TCRPX')&
    .and. hasAttribute(xCol, 'TCRVL')&
    .and. hasAttribute(xCol, 'TCDLT')&
    .and. hasAttribute(xCol, 'TCUNI'))) then
      call warning('NoWCS', 'Incomplete WCS information available for image &
      &column '//xColName)

      ctype1 = xColName
      crpix1 = 1
      crval1 = xMin + xPixelSize / 2.0
      cdelt1 = xPixelSize
      cunit1 = ''

    else
      ctype1 = stringAttribute(xCol, 'TCTYP')
      xRefPixel = 0.5 + (real64Attribute(xCol, 'TCRPX') - xMin) / xPixelSize
      crpix1 = xRefPixel
      crval1 = real64Attribute(xCol, 'TCRVL')
      cdelt1 = real64Attribute(xCol, 'TCDLT')
      cunit1 = stringAttribute(xCol, 'TCUNI')
    end if
    call release(xCol)

    yCol = column(eventTab, yColName, READ)
    if (.not. (hasAttribute(yCol, 'TCTYP')&
    .and. hasAttribute(yCol, 'TCRPX')&
    .and. hasAttribute(yCol, 'TCRVL')&
    .and. hasAttribute(yCol, 'TCDLT')&
    .and. hasAttribute(yCol, 'TCUNI'))) then
      call warning('NoWCS', 'Incomplete WCS information available for image &
      &column '//yColName)

      ctype2 = yColName
      crpix2 = 1
      crval2 = yMin + yPixelSize / 2.0
      cdelt2 = yPixelSize * yStretch
      cunit2 = ''

    else
      ctype2 = stringAttribute(yCol, 'TCTYP')
      yRefPixel = 0.5 + (real64Attribute(yCol, 'TCRPX') - yMin) / yPixelSize
      crpix2 = yRefPixel
      crval2 = real64Attribute(yCol, 'TCRVL')
      cdelt2 = real64Attribute(yCol, 'TCDLT')
      cunit2 = stringAttribute(yCol, 'TCUNI')
    end if
    call release(yCol)

    outArray = addArray(outSet, 'PRIMARY', Real32, dimVector, '', '', 0)

    call setAttribute(outArray, 'HDUCLASS', 'OGIP',&
    "Format conforms to OGIP/GSFC conventions.")

    call setAttribute(outArray, 'HDUCLAS1', 'IMAGE',&
    "File contains an Image.")

!    call setAttribute(outArray, 'HDUCLAS2', 'TOTAL',&
!    "Gross Image.")
!?????

    call setAttribute(outArray, 'HDUVERS1', '1.1.0',&
    "Version of format.")
!print *, 'set misc attributes'

    call setAttribute(outArray, 'CTYPE1', ctype1, "WCS coordinate name")
    call setAttribute(outArray, 'CRPIX1', crpix1, "WCS reference pixel")
    call setAttribute(outArray, 'CRVAL1', crval1, "WCS reference pixel value")
    call setAttribute(outArray, 'CDELT1', cdelt1, "WCS pixel size")
    call setAttribute(outArray, 'CUNIT1', cunit1, "")
    call setAttribute(outArray, 'CTYPE2', ctype2, "WCS coordinate name")
    call setAttribute(outArray, 'CRPIX2', crpix2, "WCS reference pixel")
    call setAttribute(outArray, 'CRVAL2', crval2, "WCS reference pixel value")
    call setAttribute(outArray, 'CDELT2', cdelt2, "WCS pixel size")
    call setAttribute(outArray, 'CUNIT2', cunit2, "")

!*** determine data type of image???

    arrayDataReal32 => real32Array2Data(outArray)
    arrayDataReal32 = image(1:xsize, 1:ysize)

!*** write dss???

    call release(outSet)

  end subroutine writeFitsOutput

!+++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++++

  !---------------------------------------------------------------------
  subroutine makeTestEvlist
  !---------------------------------------------------------------------

    implicit none

    type(DataSetT) ::&
      eventSet
    type(TableT)   ::&
      eventTab
    type(ColumnT) ::&
      tCol, xCol, yCol
    real(kind=double), dimension(:), pointer ::&
      tColData, xColData, yColData
    integer(kind=int32) ::&
      expectedNumEvents,&
      numEvents,&
      event
    real(kind=double), parameter ::&
      startT = 1230000.0,&
      finisT = 1240000.0,&
      loX = -2000.0,&
      hiX = 2000.0,&
      loY = -2000.0,&
      hiY = 2000.0,&
      totalRate = 10.0
    real(kind=double), dimension(:), allocatable ::&
      xs_unsorted, ys_unsorted, xs, ys, ts
    integer, dimension(:), allocatable ::&
      index
    real(kind=double) ::&
      x, y, t, meanDt, deltaT, deltaX, deltaY, tOfNextEvent, tdminT, tdmaxT,&
      tdminX, tdmaxX, tdminY, tdmaxY
!    integer :: rseed
    real, dimension(:), allocatable :: probs

!integer, parameter :: numBins = 100
!integer :: binStore(numBins), status
!real :: minBinValue, maxBinValue

    !-------------------------------------------------------------------
    ! Start:

!    rseed = 1
    deltaX = hiX - loX
    deltaY = hiY - loY

    expectedNumEvents = 2.0 * totalRate * (finisT - startT) ! an overestimate
    allocate(xs_unsorted(expectedNumEvents), ys_unsorted(expectedNumEvents))
    allocate(ts(expectedNumEvents), index(expectedNumEvents)&
    , probs(expectedNumEvents))

    call getExponentialProbabilities(probs, int(expectedNumEvents))

    numEvents = 0

    meanDt = 1.0 / totalRate
    t = startT
    do
      deltaT = meanDt * probs(numEvents + 1)
      tOfNextEvent = t + deltaT
    if (tOfNextEvent .gt. finisT) exit

!      x = myran1(rseed)
!      y = myran1(rseed)
      x = myran2()
      y = myran2()
      if ((x .gt. 0.1 .and. x .lt. 0.2 .and. y .gt. 0.8 .and. y .lt. 0.9)&
      .or. (x .gt. 0.8 .and. x .lt. 0.9 .and. y .gt. 0.1 .and. y .lt. 0.2)&
      .or. (x .gt. 0.8 .and. x .lt. 0.9 .and. y .gt. 0.8 .and. y .lt. 0.9))&
      then
    cycle
      end if

      numEvents = numEvents + 1

      t = tOfNextEvent
      xs_unsorted(numEvents) = loX + deltaX * x
      ys_unsorted(numEvents) = loY + deltaY * y
      ts(numEvents) = t
    end do

    ! Standard average rate, high-variability square:

    meanDt = 0.5 / (totalRate * 0.01)
    t = startT + (finisT - startT) * 0.5
    do
      deltaT = meanDt * probs(numEvents + 1)
      tOfNextEvent = t + deltaT
    if (tOfNextEvent .gt. finisT) exit

!      x = myran1(rseed)
!      y = myran1(rseed)
      x = myran2()
      y = myran2()

      numEvents = numEvents + 1

      t = tOfNextEvent
      xs_unsorted(numEvents) = loX + 0.8 * deltaX + 0.1 * deltaX * x
      ys_unsorted(numEvents) = loY + 0.1 * deltaY + 0.1 * deltaY * y
      ts(numEvents) = t
    end do

    ! Twice average rate, no-variability square:

    meanDt = 0.25 / (totalRate * 0.01)
    t = startT
    do
      deltaT = meanDt * probs(numEvents + 1)
      tOfNextEvent = t + deltaT
    if (tOfNextEvent .gt. finisT) exit

!      x = myran1(rseed)
!      y = myran1(rseed)
      x = myran2()
      y = myran2()

      numEvents = numEvents + 1

      t = tOfNextEvent
      xs_unsorted(numEvents) = loX + 0.1 * deltaX + 0.1 * deltaX * x
      ys_unsorted(numEvents) = loY + 0.8 * deltaY + 0.1 * deltaY * y
      ts(numEvents) = t
    end do

    ! Twice average rate, high-variability square:

    meanDt = 0.125 / (totalRate * 0.01)
    t = startT + (finisT - startT) * 0.5
    do
      deltaT = meanDt * probs(numEvents + 1)
      tOfNextEvent = t + deltaT
    if (tOfNextEvent .gt. finisT) exit

!      x = myran1(rseed)
!      y = myran1(rseed)
      x = myran2()
      y = myran2()

      numEvents = numEvents + 1

      t = tOfNextEvent
      xs_unsorted(numEvents) = loX + 0.8 * deltaX + 0.1 * deltaX * x
      ys_unsorted(numEvents) = loY + 0.8 * deltaY + 0.1 * deltaY * y
      ts(numEvents) = t
    end do

    do event = 1, expectedNumEvents
      index(event) = event
    end do

    ! Now sort the time column:

!    call quick_sort(ts(1:numEvents), index(1:numEvents)) ! a routine from
call quickSort(ts(1:numEvents), index(1:numEvents)) ! a routine from
    ! ssclib/sort_mod

    allocate(xs(numEvents))
    do event = 1, numEvents
      xs(event) = xs_unsorted(index(event))
    end do
    deallocate(xs_unsorted)
    allocate(ys(numEvents))
    do event = 1, numEvents
      ys(event) = ys_unsorted(index(event))
    end do
    deallocate(ys_unsorted)

    tdminT = minval(ts(1:numEvents))
    tdmaxT = maxval(ts(1:numEvents))
    tdminX = minval(xs(1:numEvents))
    tdmaxX = maxval(xs(1:numEvents))
    tdminY = minval(ys(1:numEvents))
    tdmaxY = maxval(ys(1:numEvents))

    ! Now store the event list
    eventSet = dataSet('test_evlist.ds', CREATE)
    eventTab = addTable(eventSet, 'EVENTS', numEvents)

    tCol = addColumn(eventTab, 'TIME', real64)
    call storeMinAndMaxColVals(tCol, 'TIME', tdminT, tdmaxT, startT, finisT)
    tColData => real64Data(tCol)
    tColData = ts(1:numEvents)
    call release(tCol)

    xCol = addColumn(eventTab, 'X', real64)
    call storeMinAndMaxColVals(xCol, 'X', tdminX, tdmaxX, loX, hiX)
    xColData => real64Data(xCol)
    xColData = xs(1:numEvents)
    call release(xCol)

    yCol = addColumn(eventTab, 'Y', real64)
    call storeMinAndMaxColVals(yCol, 'Y', tdminY, tdmaxY, loY, hiY)
    yColData => real64Data(yCol)
    yColData = ys(1:numEvents)
    call release(yCol)

    call release(eventSet)

  end subroutine makeTestEvlist

  !---------------------------------------------------------------------
  subroutine storeMinAndMaxColVals(column, colName, tdmin, tdmax&
  , tlmin, tlmax)
  !---------------------------------------------------------------------

    implicit none

    character(len=*), intent(in) :: colName
    real(kind=double), intent(in) :: tdmin, tdmax, tlmin, tlmax
    type(ColumnT), intent(in) :: column

    !-------------------------------------------------------------------
    ! Start:

    call setAttribute(column, 'TDMIN', tdmin, "Actual minimum value &
    &of the "//trim(colName)//" column.")

    call setAttribute(column, 'TDMAX', tdmax, "Actual maximum value &
    &of the "    //trim(colName)//" column.")

    call setAttribute(column, 'TLMIN', tlmin, "Legal minimum value &
    &of the "    //trim(colName)//" column.")

    call setAttribute(column, 'TLMAX', tlmax, "Legal maximum value &
    &of the "    //trim(colName)//" column.")

  end subroutine storeMinAndMaxColVals

  !---------------------------------------------------------------------
  subroutine addToBin(binStore, numBins, minBinValue, maxBinValue, arg, status)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: numBins
    real, intent(in) :: minBinValue, maxBinValue, arg
    integer, intent(inout) :: binStore(numBins)
    integer, intent(out) :: status
    integer :: bin

    !-------------------------------------------------------------------
    ! Start:

    if (arg .ge. maxBinValue) then
      status = 1
  return
    else if (arg .lt. minBinValue) then
      status = 2
  return
    end if
    status = 0

    bin = 1 + floor(numBins * ((arg - minBinValue)&
    / (maxBinValue - minBinValue)))

    binStore(bin) = binStore(bin) + 1

  end subroutine addToBin

! The following subroutine has been commented out as it is not used in the task.


  !---------------------------------------------------------------------
!  subroutine plotHistogramFromInt(binStore, numBins, minX, maxX)
!  !---------------------------------------------------------------------
!
!    implicit none
!
!    integer, intent(in) :: numBins, binStore(numBins)
!    real, intent(in) :: minX, maxX
!    real :: minBinValue, maxBinValue
!    integer :: status, pgopen, bin
!    real :: x, y, realNumBinsMinus1, xEdgeWidth, yEdgeWidth, frac
!
!    !-------------------------------------------------------------------
!    ! Start:
!
!    minBinValue = min(0.0, real(minval(binStore)))
!    maxBinValue = max(0.0, real(maxval(binStore)))
!    realNumBinsMinus1 = real(numBins - 1)
!    xEdgeWidth = 0.1 * (maxX - minX)
!    yEdgeWidth = 0.1 * (maxBinValue - minBinValue)
!
!    status = pgopen('/xs')
!    call pgsvp(0.05, 0.99, 0.05, 0.99)
!    call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth, minBinValue - yEdgeWidth&
!    , maxBinValue + yEdgeWidth)
!
!    call pgsci(2)
!    do bin = 1, numBins
!      frac = (bin - 1) / realNumBinsMinus1
!      x = minX * (1.0 - frac) + maxX * frac
!      y = real(binStore(bin))
!
!      call pgmove(x, 0.0)
!      call pgdraw(x, y)
!    end do
!    call pgsci(1)
!    call pgbox('ABCNT', 0., 0, 'ABCNT', 0., 0)
!    call pgclos
!
!  end subroutine plotHistogramFromInt

  !---------------------------------------------------------------------
  subroutine plotHistogramFromReal(yVals, numBins, minX, maxX)
  !---------------------------------------------------------------------

    implicit none

    integer, intent(in) :: numBins
    real, intent(in) :: yVals(numBins), minX, maxX
    real :: xVals(numBins)
    integer :: status, pgopen, bin
    real :: minY, maxY, x, y, realNumBinsMinus1, xEdgeWidth, yEdgeWidth, frac
    character(300) :: cmd_text

    !-------------------------------------------------------------------
    ! Start:
        
    minY = min(0.0, minval(yVals))
    maxY = max(0.0, maxval(yVals))
    realNumBinsMinus1 = real(numBins - 1)
    xEdgeWidth = 0.1 * (maxX - minX)
    yEdgeWidth = 0.1 * (maxY - minY)

!    status = pgopen('/xs')
!    call pgsvp(0.05, 0.99, 0.05, 0.99)
!    call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth&
!    , minY - yEdgeWidth, maxY + yEdgeWidth)

!    call pgsci(2)
    do bin = 1, numBins
      frac = (bin - 1) / realNumBinsMinus1
      x = minX * (1.0 - frac) + maxX * frac
      y = yVals(bin)
      xVals(bin) = x
!      call pgmove(x, 0.0)
!      call pgdraw(x, y)
    end do

    ! Saving in temporary files the values for the arrays
    open(13, file = 'yVals', status = 'new')
    write(13, *) yVals
    close(13)

    open(14, file = 'xVals', status = 'new')
    write(14, *) xVals
    close(14)

    cmd_text = "plot_hist.py"

    call execute_command_line (cmd_text)


!    call pgsci(1)
!    call pgbox('ABCNT', 0., 0, 'ABCNT', 0., 0)
!    call pgclos

  end subroutine plotHistogramFromReal


! The following subroutine has been commented out as it is not used in the task.


  !---------------------------------------------------------------------
!  subroutine plot(x, y, size)
!  !---------------------------------------------------------------------

!    implicit none

!    integer, intent(in) :: size
!    real, intent(in) :: x(size), y(size)
!    integer :: status, pgopen
!    real :: minX, maxX, minY, maxY, xEdgeWidth, yEdgeWidth!, frac

    !-------------------------------------------------------------------
    ! Start:

!    minX = min(0.0, minval(x))
!    maxX = max(0.0, maxval(x))
!    minY = min(0.0, minval(y))
!    maxY = max(0.0, maxval(y))
!    xEdgeWidth = 0.1 * (maxX - minX)
!    yEdgeWidth = 0.1 * (maxY - minY)

!    status = pgopen('ks_prob.ps/ps')
!    status = pgopen('/xs')
!    call pgsvp(0.05, 0.99, 0.05, 0.99)
!    call pgswin(minX - xEdgeWidth, maxX + xEdgeWidth&
!    , minY - yEdgeWidth, maxY + yEdgeWidth)

!    call pgsci(2)
!call pgline(size, x, y)
!!    do bin = 1, numBins
!      frac = (bin - 1) / realNumBinsMinus1
!      x = minX * (1.0 - frac) + maxX * frac
!      y = yVals(bin)

!      call pgmove(x, 0.0)
!      call pgdraw(x, y)
!!    end do
!   call pgsci(1)
!    call pgbox('ABCNT', 0., 0, 'ABCNT', 0., 0)
!    call pgclos

!  end subroutine plot

  !---------------------------------------------------------------------
  subroutine getExponentialProbabilities(probs, number)
  !---------------------------------------------------------------------
  ! Returns a vector of values which have an exponential probability
  ! distribution with a mean value of 1.0.

    implicit none

    integer, intent(in) :: number
    real, intent(out) :: probs(number)
    real :: harvest(number)
    integer :: n

    !-------------------------------------------------------------------
    ! Start:

    call random_number(harvest)
    do n = 1, number
      probs(n) = -log(1.0 - harvest(n)) ! the subtraction of the
      ! random number from 1.0 is to prevent the occurrence of zero-
      ! valued argument to log().
    end do

  end subroutine getExponentialProbabilities

  !---------------------------------------------------------------------
  function exponentialProbability(meanVal) result(val)
  !---------------------------------------------------------------------

    implicit none

    real(kind=single), intent(in) ::&
      meanVal
    real(kind=single) :: val
    integer :: rseed
real :: randomNumber

    !-------------------------------------------------------------------
    ! Start:

    rseed = 1

randomNumber = myran1(rseed)
    val = -meanVal * log(1.0 - randomNumber) ! the subtraction of the
    ! random number from 1.0 is to prevent the occurrence of zero-
    ! valued argument to log().
print *, meanVal, 1.0 - randomNumber, val
  end function exponentialProbability

  !---------------------------------------------------------------------
  real function myran1(rseed)
  !---------------------------------------------------------------------

    implicit none

    integer rseed, ia, im, iq, ir, ntab, ndiv
    real am, eps, rnmx
    parameter (ia = 16807,&
     		   im = 2147483647,&
     		   am = 1./im,&
     		   iq = 127773,&
     		   ir = 2836,&
     		 ntab = 32,&
     		 ndiv = 1 + (im - 1) / ntab,&
     		  eps = 1.2e-7,&
     		 rnmx = 1. - eps)
    integer j, k, iv(ntab), iy
    save iv, iy
    data iv /ntab * 0/, iy /0/

    if (rseed.le.0.or.iy.eq.0) then ! initialise:
      rseed = max(-rseed, 1) ! Prevents rseed = 0.
      do j = ntab + 8, 1, -1
        k = rseed / iq
        rseed = ia * (rseed - k * iq) - ir * k
        if (rseed.lt.0) rseed = rseed + im
        if (j.le.ntab) iv(j) = rseed
      end do
      iy = iv(1)
    end if

    k = rseed / iq
    rseed = ia * (rseed - k * iq) - ir * k
    if (rseed.lt.0) rseed = rseed + im
    j = 1 + iy / ndiv
    iy = iv(j)
    iv(j) = rseed
    myran1 = min(am * iy, rnmx)

  end function myran1

  !---------------------------------------------------------------------
  real function myran2()
  !---------------------------------------------------------------------

    implicit none

    integer, parameter :: max_n = 1000
    real :: harvest(max_n)
    integer :: n
    save n
    data n /0/

    !-------------------------------------------------------------------
    ! Start:

    if (n .eq. 0) then
      call random_number(harvest)
    else if (n .ge. max_n) then
      call random_number(harvest)
      n = 0
    end if
    n = n + 1
    myran2 = harvest(n)

  end function myran2

  !---------------------------------------------------------------------
  subroutine tsCheck1
  !---------------------------------------------------------------------
  ! Generates a random series of events with a variability >= 0 (variability > 0 is achieved by preventing events from occurring during periodic time gaps). The KS statistic is calculated by comparing this sequence to the elapsed time as a fraction of the occurrence time of the last event.

    implicit none

    integer, parameter :: numEvents = 20000
    integer :: event, ncycle
    real, dimension(numEvents) :: eventTimes, distance, unweightedKS, weightedKS, exptlProbs, probability
    real, parameter :: meanDt = 1.0, period = 5000.0, duty = 0.5, gapLength = duty * period
    real :: t, deltaT, realNumEvents, maxDistance

    !-------------------------------------------------------------------
    ! Start:
print *, numEvents

    eventTimes   = 0.0
    distance     = 0.0
    unweightedKS = 0.0
    weightedKS   = 0.0
    probability  = 0.0

    ! Generate a series of events, occurring randomly in time:

    call getExponentialProbabilities(exptlProbs, numEvents)
    t = 0.0
    realNumEvents = real(numEvents)
    do event = 1, numEvents
      deltaT = meanDt * exptlProbs(event)
      t = t + deltaT
      ncycle = floor(t / period)
      if (t - ncycle * period .lt. gapLength) t = t + gapLength
      eventTimes(event) = t
    end do

    ! Work through the series and calculate the KS statistic at each point:

    maxDistance = 0.0
    do event = 1, numEvents
      distance(event) = abs(event / realNumEvents - eventTimes(event)&
      / eventTimes(numEvents))
!      distance(event) = abs(event / realNumEvents - eventTimes(event)&
!      / meanDt / realNumEvents)
      if (distance(event) .gt. maxDistance) then
        maxDistance = distance(event)
!print *, event, eventTimes(event)/meanDt/realNumEvents, distance(event)
      end if
!      maxDistance = max(maxDistance, distance(event))
      unweightedKS(event) = maxDistance
      weightedKS(event) = unweightedKS(event) * sqrt(real(event))
      probability(event) = calcKSprob(weightedKS(event))
    end do

    ! Plot:

    call plotHistogramFromReal(weightedKS, numEvents, 0.0&
    , eventTimes(numEvents))

  end subroutine tsCheck1

  !---------------------------------------------------------------------
  subroutine tsCheck2
  !---------------------------------------------------------------------
  ! Generates two random series of events. There is no variability in this series. The KS statistic is calculated by comparing this sequence to the elapsed time as a fraction of the occurrence time of the last event.

    implicit none

    !-------------------------------------------------------------------
    ! Start:

  end subroutine tsCheck2

  !---------------------------------------------------------------------
  subroutine plotKsProbFn
  !---------------------------------------------------------------------

    implicit none

    integer, parameter :: max_i = 100
    integer :: i
    real, dimension(max_i) :: x, y
    real :: frac
    real, parameter :: min_x = 0.0, max_x = 2.0
real :: x_old, x_new

    !-------------------------------------------------------------------
    ! Start:

    do i = 1, max_i
      frac = (i - 1) / real(max_i - 1)
      x(i) = min_x * (1.0 - frac) + max_x * frac
!      y(i) = calcKSprob(x(i))
      y(i) = calcKSprob1der(x(i))
    end do

!   Commented out since it is not used
!    call plot(x, y, max_i)

    ! Newton's law estimation of ks at which prob = 0.5:
    x_old = 0.8
    do
      x_new = x_old + (0.5 - calcKSprob(x_old)) / calcKSprob1der(x_old)
      if (abs(x_new - x_old) .lt. 0.00001) then
    exit
      end if
      x_old = x_new
print *, x_new
    end do
print *, x_new

  end subroutine plotKsProbFn

  !---------------------------------------------------------------------
  function calcKSprob(ks) result(prob)
  !---------------------------------------------------------------------
  ! Adapted from the function probks() in Chapter 14.3 of Numerical Recipes
  ! for Fortran. It calculates
  !	   
  !	                __inf
  !	                \      j-1
  !	   prob(ks) = 2  \ (-1)   exp[-2*(j * ks)^2]
  !	                 /
  !	                /_
  !	                j=1
  !

    implicit none

    real, intent(in) :: ks
    integer :: j
    real :: prob, term, previousTerm, sum, sign, minusTwoTimesKsSquared
    integer, parameter :: max_j = 100
    logical :: converged
    real, parameter :: convCriterion1 = 0.001, convCriterion2 = 5.0e-9

    !-------------------------------------------------------------------
    ! Start:

    minusTwoTimesKsSquared = -2.0 * ks * ks
    sum = 0.0
    sign = 1.0
    previousTerm = 0.0
    converged = .false.
    do j = 1, max_j
      term = sign * exp(j * j * minusTwoTimesKsSquared)
      sum = sum + term

      if (abs(term) .le. convCriterion1 * abs(previousTerm)&
      .or. abs(term) .le. convCriterion2 * sum) then
        converged = .true.
    exit
      end if

      sign = -sign
      previousTerm = term
    end do

    if (converged) then
      prob = 2.0 * sum
    else
      prob = 1.0
    end if

  end function calcKSprob

  !---------------------------------------------------------------------
  function calcKSprob1der(ks) result(probDer)
  !---------------------------------------------------------------------
  ! This calculates the first derivative of the function probks() described in Chapter 14.3 of Numerical Recipes for Fortran, ie
  !	   
  !	                   __inf
  !	                   \      j
  !	   probDer(ks) = 8  \ (-1)  j^2 x exp[-2*(j * ks)^2]
  !	                    /
  !	                   /_
  !	                   j=1
  !

    implicit none

    real, intent(in) :: ks
    integer :: j
    real :: probDer, term, previousTerm, sum, sign, minusTwoTimesKsSquared
    integer, parameter :: max_j = 100
    logical :: converged
    real, parameter :: convCriterion1 = 0.001, convCriterion2 = 5.0e-9

    !-------------------------------------------------------------------
    ! Start:

    minusTwoTimesKsSquared = -2.0 * ks * ks
    sum = 0.0
    sign = -1.0
    previousTerm = 0.0
    converged = .false.
    do j = 1, max_j
      term = sign * real(j * j) * ks * exp(j * j * minusTwoTimesKsSquared)
      sum = sum + term

      if (abs(term) .le. convCriterion1 * abs(previousTerm)&
      .or. abs(term) .le. convCriterion2 * sum) then
        converged = .true.
    exit
      end if

      sign = -sign
      previousTerm = term
    end do

    if (converged) then
      probDer = 8.0 * sum
    else
      probDer = 0.0
    end if

  end function calcKSprob1der

  !---------------------------------------------------------------------
  subroutine generic
  !---------------------------------------------------------------------

    implicit none

    !-------------------------------------------------------------------
    ! Start:

  end subroutine generic
end module evarimgen_mod
